/**
 * \file list-kind-flags.h
 * \brief object kind flags
 *
 * These are flags that go on object kinds, egos and artifacts, but not on
 * individual objects.
 *
 * Flags below start from 1 on line 13, so a flag's sequence number is
 * its line number minus 12.
 */
/* symbol			message */
KF(NONE,			"")
KF(RAND_HI_RES,		"")
KF(RAND_SUSTAIN,	"")
KF(RAND_POWER,		"")
KF(INSTA_ART,		"")
KF(QUEST_ART,		"")
KF(EASY_KNOW,		"")
KF(GOOD,			"")
KF(SHOW_DICE,		"")
KF(SHOW_MULT,		"")
KF(SHOOTS_SHOTS,	"")
KF(SHOOTS_ARROWS,	"")
KF(SHOOTS_BOLTS,	"")
KF(RAND_BASE_RES,	"")
KF(RAND_RES_POWER,	"")
KF(MAX,				"")
