/**
 * \file list-randart-properties.h
 * \brief List of properties used in randart generation
 */
/* string		 			supercharge */
ART_IDX(BOW_SHOTS,			false)
ART_IDX(BOW_MIGHT,			false)
ART_IDX(BOW_BRAND,			false)
ART_IDX(BOW_SLAY,			false)
ART_IDX(WEAPON_HIT,			false)
ART_IDX(WEAPON_DAM,			false)
ART_IDX(NONWEAPON_HIT,		false)
ART_IDX(NONWEAPON_DAM,		false)
ART_IDX(NONWEAPON_HIT_DAM,	false)
ART_IDX(NONWEAPON_BRAND,	false)
ART_IDX(NONWEAPON_SLAY,		false)
ART_IDX(NONWEAPON_BLOWS,	false)
ART_IDX(NONWEAPON_SHOTS,	false)
ART_IDX(MELEE_BLESS,		false)
ART_IDX(MELEE_BRAND,		false)
ART_IDX(MELEE_SLAY,			false)
ART_IDX(MELEE_SINV,			false)
ART_IDX(MELEE_BLOWS,		false)
ART_IDX(MELEE_AC,			false)
ART_IDX(MELEE_DICE,			false)
ART_IDX(MELEE_WEIGHT,		false)
ART_IDX(MELEE_TUNN,			false)
ART_IDX(ALLARMOR_WEIGHT,	false)
ART_IDX(BOOT_AC,			false)
ART_IDX(BOOT_FEATHER,		false)
ART_IDX(BOOT_STEALTH,		false)
ART_IDX(BOOT_TRAP_IMM,		false)
ART_IDX(BOOT_SPEED,			false)
ART_IDX(BOOT_MOVES,			false)
ART_IDX(GLOVE_AC,			false)
ART_IDX(GLOVE_FA,			false)
ART_IDX(GLOVE_DEX,			false)
ART_IDX(GLOVE_HIT_DAM,		false)
ART_IDX(HELM_AC,			false)
ART_IDX(HELM_RBLIND,		false)
ART_IDX(HELM_ESP,			false)
ART_IDX(HELM_SINV,			false)
ART_IDX(HELM_WIS,			false)
ART_IDX(HELM_INT,			false)
ART_IDX(SHIELD_AC,			false)
ART_IDX(SHIELD_LRES,		false)
ART_IDX(CLOAK_AC,			false)
ART_IDX(CLOAK_STEALTH,		false)
ART_IDX(ARMOR_AC,			false)
ART_IDX(ARMOR_STEALTH,		false)
ART_IDX(ARMOR_HLIFE,		false)
ART_IDX(ARMOR_CON,			false)
ART_IDX(ARMOR_LRES,			false)
ART_IDX(ARMOR_ALLRES,		false)
ART_IDX(ARMOR_HRES,			false)
ART_IDX(GEN_STAT,			false)
ART_IDX(GEN_SUST,			false)
ART_IDX(GEN_STEALTH,		false)
ART_IDX(GEN_SEARCH,			false)
ART_IDX(GEN_INFRA,			false)
ART_IDX(GEN_SPEED,			false)
ART_IDX(GEN_IMMUNE,			false)
ART_IDX(GEN_FA,				false)
ART_IDX(GEN_HLIFE,			false)
ART_IDX(GEN_FEATHER,		false)
ART_IDX(GEN_LIGHT,			false)
ART_IDX(GEN_SINV,			false)
ART_IDX(GEN_ESP,			false)
ART_IDX(GEN_SDIG,			false)
ART_IDX(GEN_REGEN,			false)
ART_IDX(GEN_LRES,			false)
ART_IDX(GEN_RPOIS,			false)
ART_IDX(GEN_RFEAR,			false)
ART_IDX(GEN_RLIGHT,			false)
ART_IDX(GEN_RDARK,			false)
ART_IDX(GEN_RBLIND,			false)
ART_IDX(GEN_RCONF,			false)
ART_IDX(GEN_RSOUND,			false)
ART_IDX(GEN_RSHARD,			false)
ART_IDX(GEN_RNEXUS,			false)
ART_IDX(GEN_RNETHER,		false)
ART_IDX(GEN_RCHAOS,			false)
ART_IDX(GEN_RDISEN,			false)
ART_IDX(GEN_AC,				false)
ART_IDX(GEN_TUNN,			false)
ART_IDX(GEN_ACTIV,			false)
ART_IDX(GEN_PSTUN,			false)
ART_IDX(GEN_DAM_RED,		false)
ART_IDX(GEN_MOVES,			false)
ART_IDX(GEN_TRAP_IMM,		false)
ART_IDX(WEAPON_AGGR,		false)
ART_IDX(NONWEAPON_AGGR,		false)
ART_IDX(MELEE_DICE_SUPER,	true)
ART_IDX(BOW_SHOTS_SUPER,	true)
ART_IDX(BOW_MIGHT_SUPER,	true)
ART_IDX(GEN_SPEED_SUPER,	true)
ART_IDX(MELEE_BLOWS_SUPER,	true)
ART_IDX(MELEE_AC_SUPER,		true)
ART_IDX(GEN_AC_SUPER,		true)
ART_IDX(TOTAL,				false)
