#ifndef __DLMALLOC_CONFIG_H__
#define __DLMALLOC_CONFIG_H__

#define INSECURE		0
#define ONLY_MSPACES    	1
#define HAVE_MORECORE		0
#define HAVE_MMAP		0
#define MSPACES			1
#define LACKS_SYS_MMAN_H 	1
#define LACKS_SCHED_H		1
#define FOOTERS			0
#define NO_MALLOC_STATS		1
#define PROCEED_ON_ERROR	1
#define malloc_getpagesize	4096

#endif
