/**
 * \file ui-history.h
 * \brief Character auto-history display
 *
 * Copyright (c) 2007 J.D. White
 *
 * This work is free software; you can redistribute it and/or modify it
 * under the terms of either:
 *
 * a) the GNU General Public License as published by the Free Software
 *    Foundation, version 2, or
 *
 * b) the "Angband licence":
 *    This software may be copied and distributed for educational, research,
 *    and not for profit purposes provided that this copyright and statement
 *    are included in all such copies.  Other copyrights may also apply.
 */

#ifndef UI_HISTORY_H
#define UI_HISTORY_H

void history_display(void);
void dump_history(ang_file *file);

#endif /* !UI_HISTORY_H */
