/**
 * \file scrnshot.h
 * \brief Prototypes for functions from scrnshot.c
 */

#ifndef INCLUDED_WIN_SCRNSHOT_H
#define INCLUDED_WIN_SCRNSHOT_H

bool SaveWindow_PNG(HWND hWnd, LPSTR lpFileName);

#endif /* INCLUDED_WIN_SCRNSHOT_H */
