/**
 * \file ui-entry-init.h
 * \brief Declare parsers for the ui-entry configuration
 */
#ifndef INCLUDED_UI_ENTRY_INIT_H
#define INCLUDED_UI_ENTRY_INIT_H

#include "datafile.h"

/* From ui-entry.c */
extern struct file_parser ui_entry_parser;
/* From ui-entry-renderers.c */
extern struct file_parser ui_entry_renderer_parser;

#endif /* INCLUDED_UI_ENTRY_INIT_H */
