/**
 * \file snd-cocoa.h
 * \brief Sound handling for the Cocoa frontend on macOS
 */

#ifndef INCLUDED_SND_COCOA_H
#define INCLUDED_SND_COCOA_H

errr init_sound_cocoa(struct sound_hooks *hooks, int argc, char **argv);

#endif /* !INCLUDED_SND_COCOA_H */
