/**
 * \file list-dun-profiles.h
 * \brief matches dungeon profile names to their building functions
 */

DUN("town", town)
DUN("modified", modified)
DUN("moria", moria)
DUN("lair", lair)
DUN("gauntlet", gauntlet)
DUN("hard centre", hard_centre)
DUN("labyrinth", labyrinth)
DUN("cavern", cavern)
DUN("classic", classic)
