/**
 * \file list-elements.h
 * \brief Elements used in spells and other attacks.
 */
ELEM(ACID)
ELEM(ELEC)
ELEM(FIRE)
ELEM(COLD)
ELEM(POIS)
ELEM(LIGHT)
ELEM(DARK)
ELEM(SOUND)
ELEM(SHARD)
ELEM(NEXUS)
ELEM(NETHER)
ELEM(CHAOS)
ELEM(DISEN)
ELEM(WATER)
ELEM(ICE)
ELEM(GRAVITY)
ELEM(INERTIA)
ELEM(FORCE)
ELEM(TIME)
ELEM(PLASMA)
ELEM(METEOR)
ELEM(MISSILE)
ELEM(MANA)
ELEM(HOLY_ORB)
ELEM(ARROW)
