/**
 * \file list-mon-temp-flags.h
 * \brief Temporary monster flags
 */

MFLAG(NONE,		"")
MFLAG(VIEW,		"Monster is in line of sight")
MFLAG(ACTIVE,	"Monster is in active mode")
MFLAG(NICE,		"Monster is still being nice")
MFLAG(SHOW,		"Monster is recently memorized")
MFLAG(MARK,		"Monster is currently memorized")
MFLAG(VISIBLE,	"Monster is \"visible\"")
MFLAG(CAMOUFLAGE,"Player doesn't know this is a monster")
MFLAG(AWARE,	"Monster is aware of the player")
MFLAG(HANDLED,	"Monster has been processed this turn")
MFLAG(TRACKING,	"Monster is tracking the player by sound or scent")
