/**
 * \file list-message.h
 * \brief List of message types
 *
 * sound - the name, case-insensitive, of the entry in sound.prf (redundant)
 */

/* id					sound */
MSG(GENERIC,			"")
MSG(BIRTH,				"")
MSG(HIT,				"hit")
MSG(MISS,				"miss")
MSG(FLEE,				"flee")
MSG(DROP,				"drop")
MSG(KILL,				"kill")
MSG(LEVEL,				"level")
MSG(DEATH,				"death")
MSG(STUDY,				"study")
MSG(TELEPORT,			"teleport")
MSG(SHOOT,				"shoot")
MSG(QUAFF,				"quaff")
MSG(ZAP_ROD,			"zap_rod")
MSG(WALK,				"walk")
MSG(TPOTHER,			"tpother")
MSG(HITWALL,			"hitwall")
MSG(EAT,				"eat")
MSG(STORE1,				"store1")
MSG(STORE2,				"store2")
MSG(STORE3,				"store3")
MSG(STORE4,				"store4")
MSG(DIG,				"dig")
MSG(OPENDOOR,			"opendoor")
MSG(SHUTDOOR,			"shutdoor")
MSG(TPLEVEL,			"tplevel")
MSG(BELL,				"bell")
MSG(NOTHING_TO_OPEN,	"nothing_to_open")
MSG(LOCKPICK_FAIL,		"lockpick_fail")
MSG(STAIRS_DOWN,		"stairs_down")
MSG(HITPOINT_WARN,		"hitpoint_warn")
MSG(ACT_ARTIFACT,		"act_artifact")
MSG(USE_STAFF,			"use_staff")
MSG(DESTROY,			"destroy")
MSG(MON_HIT,			"mon_hit")
MSG(MON_TOUCH,			"mon_touch")
MSG(MON_PUNCH,			"mon_punch")
MSG(MON_KICK,			"mon_kick")
MSG(MON_CLAW,			"mon_claw")
MSG(MON_BITE,			"mon_bite")
MSG(MON_STING,			"mon_sting")
MSG(MON_BUTT,			"mon_butt")
MSG(MON_CRUSH,			"mon_crush")
MSG(MON_ENGULF,			"mon_engulf")
MSG(MON_CRAWL,			"mon_crawl")
MSG(MON_DROOL,			"mon_drool")
MSG(MON_SPIT,			"mon_spit")
MSG(MON_GAZE,			"mon_gaze")
MSG(MON_WAIL,			"mon_wail")
MSG(MON_SPORE,			"mon_spore")
MSG(MON_BEG,			"mon_beg")
MSG(MON_INSULT,			"mon_insult")
MSG(MON_MOAN,			"mon_moan")
MSG(RECOVER,			"recover")
MSG(BLIND,				"blind")
MSG(CONFUSED,			"confused")
MSG(POISONED,			"poisoned")
MSG(AFRAID,				"afraid")
MSG(PARALYZED,			"paralyzed")
MSG(DRUGGED,			"drugged")
MSG(SPEED,				"speed")
MSG(SLOW,				"slow")
MSG(SHIELD,				"shield")
MSG(BLESSED,			"blessed")
MSG(HERO,				"hero")
MSG(BERSERK,			"berserk")
MSG(BOLD,				"bold")
MSG(PROT_EVIL,			"prot_evil")
MSG(INVULN,				"invuln")
MSG(SEE_INVIS,			"see_invis")
MSG(INFRARED,			"infrared")
MSG(RES_ACID,			"res_acid")
MSG(RES_ELEC,			"res_elec")
MSG(RES_FIRE,			"res_fire")
MSG(RES_COLD,			"res_cold")
MSG(RES_POIS,			"res_pois")
MSG(STUN,				"stun")
MSG(CUT,				"cut")
MSG(STAIRS_UP,			"stairs_up")
MSG(STORE_ENTER,		"store_enter")
MSG(STORE_LEAVE,		"store_leave")
MSG(STORE_HOME,			"store_home")
MSG(MONEY1,				"money1")
MSG(MONEY2,				"money2")
MSG(MONEY3,				"money3")
MSG(SHOOT_HIT,			"shoot_hit")
MSG(STORE5,				"store5")
MSG(LOCKPICK,			"lockpick")
MSG(DISARM,				"disarm")
MSG(IDENT_BAD,			"identify_bad")
MSG(IDENT_EGO,			"identify_ego")
MSG(IDENT_ART,			"identify_art")
MSG(BR_ELEMENTS,		"breathe_elements")
MSG(BR_FROST,			"breathe_frost")
MSG(BR_ELEC,			"breathe_elec")
MSG(BR_ACID,			"breathe_acid")
MSG(BR_GAS,				"breathe_gas")
MSG(BR_FIRE,			"breathe_fire")
MSG(BR_DISEN,			"breathe_disenchant")
MSG(BR_CHAOS,			"breathe_chaos")
MSG(BR_SHARDS,			"breathe_shards")
MSG(BR_SOUND,			"breathe_sound")
MSG(BR_LIGHT,			"breathe_light")
MSG(BR_DARK,			"breathe_dark")
MSG(BR_NETHER,			"breathe_nether")
MSG(BR_NEXUS,			"breathe_nexus")
MSG(BR_TIME,			"breathe_time")
MSG(BR_INERTIA,			"breathe_inertia")
MSG(BR_GRAVITY,			"breathe_gravity")
MSG(BR_PLASMA,			"breathe_plasma")
MSG(BR_FORCE,			"breathe_force")
MSG(SUM_MONSTER,		"summon_monster")
MSG(SUM_AINU,			"summon_ainu")
MSG(SUM_UNDEAD,			"summon_undead")
MSG(SUM_ANIMAL,			"summon_animal")
MSG(SUM_SPIDER,			"summon_spider")
MSG(SUM_HOUND,			"summon_hound")
MSG(SUM_HYDRA,			"summon_hydra")
MSG(SUM_DEMON,			"summon_demon")
MSG(SUM_DRAGON,			"summon_dragon")
MSG(SUM_HI_UNDEAD,		"summon_gr_undead")
MSG(SUM_HI_DRAGON,		"summon_gr_dragon")
MSG(SUM_HI_DEMON,		"summon_gr_demon")
MSG(SUM_WRAITH,			"summon_ringwraith")
MSG(SUM_UNIQUE,			"summon_unique")
MSG(WIELD,				"wield")
MSG(QUIVER,				"quiver")
MSG(CURSED,				"cursed")
MSG(RUNE,				"rune")
MSG(HUNGRY,				"hungry")
MSG(NOTICE,				"notice")
MSG(AMBIENT_DAY,		"ambient_day")
MSG(AMBIENT_NITE,		"ambient_nite")
MSG(AMBIENT_DNG1,		"ambient_dng1")
MSG(AMBIENT_DNG2,		"ambient_dng2")
MSG(AMBIENT_DNG3,		"ambient_dng3")
MSG(AMBIENT_DNG4,		"ambient_dng4")
MSG(AMBIENT_DNG5,		"ambient_dng5")
MSG(CREATE_TRAP,		"mon_create_trap")
MSG(SHRIEK,				"mon_shriek")
MSG(CAST_FEAR,			"mon_cast_fear")
MSG(HIT_GOOD,			"hit_good")
MSG(HIT_GREAT,			"hit_great")
MSG(HIT_SUPERB,			"hit_superb")
MSG(HIT_HI_GREAT,		"hit_hi_great")
MSG(HIT_HI_SUPERB,		"hit_hi_superb")
MSG(SPELL,				"cast_spell")
MSG(PRAYER,				"pray_prayer")
MSG(KILL_UNIQUE,		"kill_unique")
MSG(KILL_KING,			"kill_king")
MSG(DRAIN_STAT,			"drain_stat")
MSG(MULTIPLY,			"multiply")
MSG(SCRAMBLE,			"scramble")
MSG(MAX,				NULL)
