/**
 * \file list-projections.h
 * \brief Spell types used by project() and related functions.
 *
 * ===== NOTICE =====
 * If you add new entries to this file, you must
 * also update lib/gamedata/projection.txt!
 * ===== NOTICE =====
 */
PROJ(LIGHT_WEAK)
PROJ(DARK_WEAK)
PROJ(KILL_WALL)
PROJ(KILL_DOOR)
PROJ(KILL_TRAP)
PROJ(MAKE_DOOR)
PROJ(MAKE_TRAP)
PROJ(AWAY_UNDEAD)
PROJ(AWAY_SPIRIT)
PROJ(AWAY_EVIL)
PROJ(AWAY_ALL)
PROJ(TURN_UNDEAD)
PROJ(TURN_EVIL)
PROJ(TURN_LIVING)
PROJ(TURN_ALL)
PROJ(DISP_UNDEAD)
PROJ(DISP_EVIL)
PROJ(DISP_ALL)
PROJ(SLEEP_UNDEAD)
PROJ(SLEEP_EVIL)
PROJ(SLEEP_ALL)
PROJ(MON_CLONE)
PROJ(MON_POLY)
PROJ(MON_HEAL)
PROJ(MON_SPEED)
PROJ(MON_SLOW)
PROJ(MON_CONF)
PROJ(MON_HOLD)
PROJ(MON_STUN)
PROJ(MON_DRAIN)
PROJ(MON_CRUSH)
