/**
 * \file list-ignore-types.h
 * \brief types of object used for ignoring by quality or ego
 */

/* index					description */
ITYPE(NONE,					"")
ITYPE(SHARP,				"Sharp Melee Weapons")
ITYPE(BLUNT,				"Blunt Melee Weapons")
ITYPE(GREAT,				"Great Weapons")
ITYPE(SLING,				"Slings")
ITYPE(BOW,					"Bows")
ITYPE(CROSSBOW,				"Crossbows")
ITYPE(SHOT,					"Shots and Pebbles")
ITYPE(ARROW,				"Arrows")
ITYPE(BOLT,					"Bolts")
ITYPE(ROBE,					"Robes")
ITYPE(BODY_ARMOR,			"Body Armor")
ITYPE(BASIC_DRAGON_ARMOR,	"Basic Dragon Scale Mail")
ITYPE(MULTI_DRAGON_ARMOR,	"Multi-Hued Dragon Scale Mail")
ITYPE(HIGH_DRAGON_ARMOR,	"High Dragon Scale Mail")
ITYPE(BALANCE_DRAGON_ARMOR,	"Balance Dragon Scale Mail")
ITYPE(POWER_DRAGON_ARMOR,	"Power Dragon Scale Mail")
ITYPE(CLOAK,				"Cloaks")
ITYPE(ELVEN_CLOAK,			"Elven Cloaks")
ITYPE(SHIELD,				"Shields")
ITYPE(HEADGEAR,				"Headgear")
ITYPE(HANDGEAR,				"Handgear")
ITYPE(FEET,					"Footgear")
ITYPE(DIGGER,				"Diggers")
ITYPE(RING,					"Rings")
ITYPE(AMULET,				"Amulets")
ITYPE(LIGHT, 				"Lights")
