/**
 * \file save-charoutput.h
 * \brief Declare interface for saving CharOutput.txt for angband.live
 */
#ifndef INCLUDED_SAVE_CHAROUTPUT_H
#define INCLUDED_SAVE_CHAROUTPUT_H

#include "h-basic.h"

bool save_charoutput(void);

#endif /* INCLUDED_SAVE_CHAROUTPUT_H */

