/* File: angband.rc */
#include "win-menu.h"

ANGBAND MENU
{
	POPUP "&File"
	{
		MENUITEM "&New\tCtrl+N", IDM_FILE_NEW
		MENUITEM "&Open...\tCtrl+O", IDM_FILE_OPEN
		MENUITEM SEPARATOR
		MENUITEM "&Save\tCtrl+S", IDM_FILE_SAVE
		MENUITEM SEPARATOR
		MENUITEM "E&xit\tCtrl+X", IDM_FILE_EXIT
	}

	POPUP "&Window"
	{
		POPUP "&Visibility"
		{
			MENUITEM "Term-0 window", IDM_WINDOW_VIS_0
			MENUITEM "Term-1 window", IDM_WINDOW_VIS_1
			MENUITEM "Term-2 window", IDM_WINDOW_VIS_2
			MENUITEM "Term-3 window", IDM_WINDOW_VIS_3
			MENUITEM "Term-4 window", IDM_WINDOW_VIS_4
			MENUITEM "Term-5 window", IDM_WINDOW_VIS_5
			MENUITEM "Term-6 window", IDM_WINDOW_VIS_6
			MENUITEM "Term-7 window", IDM_WINDOW_VIS_7
		}

		POPUP "&Font"
		{
			MENUITEM "Term-0 window", IDM_WINDOW_FONT_0
			MENUITEM "Term-1 window", IDM_WINDOW_FONT_1
			MENUITEM "Term-2 window", IDM_WINDOW_FONT_2
			MENUITEM "Term-3 window", IDM_WINDOW_FONT_3
			MENUITEM "Term-4 window", IDM_WINDOW_FONT_4
			MENUITEM "Term-5 window", IDM_WINDOW_FONT_5
			MENUITEM "Term-6 window", IDM_WINDOW_FONT_6
			MENUITEM "Term-7 window", IDM_WINDOW_FONT_7
		}

		MENUITEM SEPARATOR

		POPUP "Bizarre Display"
		{
			MENUITEM "Term-0 window", IDM_WINDOW_BIZ_0
			MENUITEM "Term-1 window", IDM_WINDOW_BIZ_1
			MENUITEM "Term-2 window", IDM_WINDOW_BIZ_2
			MENUITEM "Term-3 window", IDM_WINDOW_BIZ_3
			MENUITEM "Term-4 window", IDM_WINDOW_BIZ_4
			MENUITEM "Term-5 window", IDM_WINDOW_BIZ_5
			MENUITEM "Term-6 window", IDM_WINDOW_BIZ_6
			MENUITEM "Term-7 window", IDM_WINDOW_BIZ_7
		}

		POPUP "Increase Tile Width"
		{
			MENUITEM "Term-0 window", IDM_WINDOW_I_WID_0
			MENUITEM "Term-1 window", IDM_WINDOW_I_WID_1
			MENUITEM "Term-2 window", IDM_WINDOW_I_WID_2
			MENUITEM "Term-3 window", IDM_WINDOW_I_WID_3
			MENUITEM "Term-4 window", IDM_WINDOW_I_WID_4
			MENUITEM "Term-5 window", IDM_WINDOW_I_WID_5
			MENUITEM "Term-6 window", IDM_WINDOW_I_WID_6
			MENUITEM "Term-7 window", IDM_WINDOW_I_WID_7
		}

		POPUP "Decrease Tile Width"
		{
			MENUITEM "Term-0 window", IDM_WINDOW_D_WID_0
			MENUITEM "Term-1 window", IDM_WINDOW_D_WID_1
			MENUITEM "Term-2 window", IDM_WINDOW_D_WID_2
			MENUITEM "Term-3 window", IDM_WINDOW_D_WID_3
			MENUITEM "Term-4 window", IDM_WINDOW_D_WID_4
			MENUITEM "Term-5 window", IDM_WINDOW_D_WID_5
			MENUITEM "Term-6 window", IDM_WINDOW_D_WID_6
			MENUITEM "Term-7 window", IDM_WINDOW_D_WID_7
		}

		POPUP "Increase Tile Height"
		{
			MENUITEM "Term-0 window", IDM_WINDOW_I_HGT_0
			MENUITEM "Term-1 window", IDM_WINDOW_I_HGT_1
			MENUITEM "Term-2 window", IDM_WINDOW_I_HGT_2
			MENUITEM "Term-3 window", IDM_WINDOW_I_HGT_3
			MENUITEM "Term-4 window", IDM_WINDOW_I_HGT_4
			MENUITEM "Term-5 window", IDM_WINDOW_I_HGT_5
			MENUITEM "Term-6 window", IDM_WINDOW_I_HGT_6
			MENUITEM "Term-7 window", IDM_WINDOW_I_HGT_7
		}

		POPUP "Decrease Tile Height"
		{
			MENUITEM "Term-0 window", IDM_WINDOW_D_HGT_0
			MENUITEM "Term-1 window", IDM_WINDOW_D_HGT_1
			MENUITEM "Term-2 window", IDM_WINDOW_D_HGT_2
			MENUITEM "Term-3 window", IDM_WINDOW_D_HGT_3
			MENUITEM "Term-4 window", IDM_WINDOW_D_HGT_4
			MENUITEM "Term-5 window", IDM_WINDOW_D_HGT_5
			MENUITEM "Term-6 window", IDM_WINDOW_D_HGT_6
			MENUITEM "Term-7 window", IDM_WINDOW_D_HGT_7
		}

		MENUITEM SEPARATOR

		POPUP "Term 0 Font Tile Size"
		{
			MENUITEM "Font",    IDM_TILE_FONT
			MENUITEM "8 x 16",  IDM_TILE_08X16
			MENUITEM "8 x 8",   IDM_TILE_08X08
			MENUITEM "16 x 16", IDM_TILE_16X16
			MENUITEM "32 x 32", IDM_TILE_32X32
			MENUITEM "16 x 32", IDM_TILE_16X32
			MENUITEM "12 x 20", IDM_TILE_12X20
			MENUITEM "16 x 25", IDM_TILE_16X25
		}
		MENUITEM SEPARATOR
		MENUITEM "Term Options",    IDM_WINDOW_OPT
		MENUITEM "Reset Layout",    IDM_WINDOW_RESET
	}

	POPUP "&Options"
	{
		POPUP "&Graphics"
		{
			MENUITEM "&None", IDM_OPTIONS_GRAPHICS_NONE
			MENUITEM SEPARATOR
			MENUITEM "Enable nice &graphics", IDM_OPTIONS_GRAPHICS_NICE
			MENUITEM SEPARATOR
			POPUP "&Tile Multiplier"
			{
				MENUITEM "&1x1", IDM_OPTIONS_TILE_1x1
				MENUITEM "&2x1", IDM_OPTIONS_TILE_2x1
				MENUITEM "2x2", IDM_OPTIONS_TILE_2x2
				MENUITEM "3x1", IDM_OPTIONS_TILE_3x1
				MENUITEM "&3x3", IDM_OPTIONS_TILE_3x3
				MENUITEM "&4x2", IDM_OPTIONS_TILE_4x2
				MENUITEM "4x4", IDM_OPTIONS_TILE_4x4
				MENUITEM "&6x3", IDM_OPTIONS_TILE_6x3
				MENUITEM "6x6", IDM_OPTIONS_TILE_6x6
				MENUITEM "&8x4", IDM_OPTIONS_TILE_8x4
				MENUITEM "8x8", IDM_OPTIONS_TILE_8x8
				MENUITEM "16x8", IDM_OPTIONS_TILE_16x8
				MENUITEM "16x16", IDM_OPTIONS_TILE_16x16
			}
		}

		MENUITEM SEPARATOR
		MENUITEM "Low Priority", IDM_OPTIONS_LOW_PRIORITY
		MENUITEM "Activate Screensaver", IDM_OPTIONS_SAVER
		MENUITEM "&Screenshot", IDM_OPTIONS_SCREENSHOT
	}

	POPUP "&Help"
	{
		MENUITEM "&Contents", IDM_HELP_GENERAL
	}
}

ANGBAND ICON "angband.ico"

